@@ -15,22 +15,26 @@
 # GNU General Public License for more details.
 
 # You should have received a copy of the GNU General Public License
-# along with this program; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+# along with this program; if not, see <http://www.gnu.org/licenses/>.
+from __future__ import absolute_import
 
-from mercurial import util
-from svn.core import SubversionException, Pool
-import svn.ra
 import svn.client
 import svn.core
+import svn.ra
+
+Pool = svn.core.Pool
+SubversionException = svn.core.SubversionException
+
+from mercurial import (
+    util,
+)
 
 # Some older versions of the Python bindings need to be
 # explicitly initialized. But what we want to do probably
 # won't work worth a darn against those libraries anyway!
 svn.ra.initialize()
 
-svn_config = svn.core.svn_config_get_config(None)
-
+svn_config = None
 
 def _create_auth_baton(pool):
     """Create a Subversion authentication baton. """
@@ -44,7 +48,7 @@ def _create_auth_baton(pool):
         svn.client.get_ssl_client_cert_pw_file_provider(pool),
         svn.client.get_ssl_server_trust_file_provider(pool),
         ]
-    # Platform-dependant authentication methods
+    # Platform-dependent authentication methods
     getprovider = getattr(svn.core, 'svn_auth_get_platform_specific_provider',
                           None)
     if getprovider:
@@ -77,18 +81,17 @@ class SvnRaTransport(object):
         if ra is None or not util.safehasattr(svn.ra, 'reparent'):
             self.client = svn.client.create_context(self.pool)
             ab = _create_auth_baton(self.pool)
-            if False:
-                svn.core.svn_auth_set_parameter(
-                    ab, svn.core.SVN_AUTH_PARAM_DEFAULT_USERNAME, self.username)
-                svn.core.svn_auth_set_parameter(
-                    ab, svn.core.SVN_AUTH_PARAM_DEFAULT_PASSWORD, self.password)
             self.client.auth_baton = ab
+            global svn_config
+            if svn_config is None:
+                svn_config = svn.core.svn_config_get_config(None)
             self.client.config = svn_config
             try:
                 self.ra = svn.client.open_ra_session(
-                    self.svn_url.encode('utf8'),
+                    self.svn_url,
                     self.client, self.pool)
-            except SubversionException, (inst, num):
+            except SubversionException as xxx_todo_changeme:
+                (inst, num) = xxx_todo_changeme.args
                 if num in (svn.core.SVN_ERR_RA_ILLEGAL_URL,
                            svn.core.SVN_ERR_RA_LOCAL_REPOS_OPEN_FAILED,
                            svn.core.SVN_ERR_BAD_URL):
